@@ -36,7 +36,6 @@ class PassRegistry;
 class raw_ostream;
 
 void initializeIslAstInfoWrapperPassPass(PassRegistry &);
-
 } // namespace llvm
 
 struct isl_ast_build;
@@ -104,6 +103,10 @@ public:
     /// Cleanup all isl structs on destruction.
     ~IslAstUserPayload();
 
+    /// Does the dependence analysis determine that there are no loop-carried
+    /// dependencies?
+    bool IsParallel = false;
+
     /// Flag to mark innermost loops.
     bool IsInnermost = false;
 
@@ -117,7 +120,7 @@ public:
     bool IsReductionParallel = false;
 
     /// The minimal dependence distance for non parallel loops.
-    isl_pw_aff *MinimalDependenceDistance = nullptr;
+    isl::pw_aff MinimalDependenceDistance;
 
     /// The build environment at the time this node was constructed.
     isl_ast_build *Build = nullptr;
@@ -230,7 +233,6 @@ struct IslAstPrinterPass : public PassInfoMixin<IslAstPrinterPass> {
 
   raw_ostream &OS;
 };
-
 } // namespace polly
 
 #endif // POLLY_ISLAST_H
